/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

import React from "react";
import PropTypes from "prop-types";
import classNames from "classnames";
import { withStyles } from "@material-ui/core/styles";
import CssBaseline from "@material-ui/core/CssBaseline";
import Drawer from "@material-ui/core/Drawer";
import AppBar from "@material-ui/core/AppBar";
import Toolbar from "@material-ui/core/Toolbar";
import Typography from "@material-ui/core/Typography";
import Divider from "@material-ui/core/Divider";
import IconButton from "@material-ui/core/IconButton";
import MenuIcon from "@material-ui/icons/Menu";
import ExitToAppIcon from "@material-ui/icons/ExitToApp";
import UserServer from "./UserServer";
import UserServerList from "./UserServerList";
import GroupList from "./GroupList";
import UserList from "./UserList";
import ServerList from "./ServerList";
import AdminMenu from "./AdminMenu";
import UserDetails from "./UserDetails";
import PublicKey from "./PublicKey";
import AccountCircleIcon from "@material-ui/icons/AccountCircle";
import { logout } from "./util/api";

const drawerWidth = 240;

const styles = (theme) => ({
  root: {
    display: "flex",
  },
  toolbar: {
    paddingRight: 24, // keep right padding when drawer closed
  },
  toolbarIcon: {
    display: "flex",
    alignItems: "center",
    justifyContent: "flex-end",
    padding: "0 8px",
    ...theme.mixins.toolbar,
  },
  appBar: {
    zIndex: theme.zIndex.drawer + 1,
    transition: theme.transitions.create(["width", "margin"], {
      easing: theme.transitions.easing.sharp,
      duration: theme.transitions.duration.leavingScreen,
    }),
  },
  appBarShift: {
    marginLeft: drawerWidth,
    width: `calc(100% - ${drawerWidth}px)`,
    transition: theme.transitions.create(["width", "margin"], {
      easing: theme.transitions.easing.sharp,
      duration: theme.transitions.duration.enteringScreen,
    }),
  },
  menuButton: {
    marginLeft: 12,
    marginRight: 36,
  },
  menuButtonHidden: {
    display: "none",
  },
  title: {
    flexGrow: 1,
  },
  drawerPaper: {
    position: "relative",
    whiteSpace: "nowrap",
    width: drawerWidth,
    transition: theme.transitions.create("width", {
      easing: theme.transitions.easing.sharp,
      duration: theme.transitions.duration.enteringScreen,
    }),
  },
  drawerPaperClose: {
    overflowX: "hidden",
    transition: theme.transitions.create("width", {
      easing: theme.transitions.easing.sharp,
      duration: theme.transitions.duration.leavingScreen,
    }),
    width: theme.spacing.unit * 7,
    [theme.breakpoints.up("sm")]: {
      width: theme.spacing.unit * 9,
    },
  },
  appBarSpacer: theme.mixins.toolbar,
  content: {
    flexGrow: 1,
    padding: theme.spacing.unit * 3,
    height: "100vh",
    overflow: "auto",
  },
  chartContainer: {
    marginLeft: -22,
  },
  tableContainer: {
    height: 320,
  },
});

class Dashboard extends React.Component {
  state = {
    open: true,
    activePage: null,
    activeServer: 1,
    activeServerName: null,
  };

  handleDrawerOpen = () => {
    this.setState({ open: true });
  };

  handleDrawerClose = () => {
    this.setState({ open: false });
  };

  handleLogout = () => {
    logout().then((json) => {
      this.props.setLoggedOut();
    });
  };

  setServer = (server_id, server_name) => {
    this.setState({
      activePage: "server",
      activeServer: server_id,
      activeServerName: server_name,
    });
  };

  setActivePage = (page) => {
    this.setState({ activePage: page });
  };

  page = (page) => {
    // Show the right component
    switch (page) {
      case "server":
        return (
          <UserServer
            serverID={this.state.activeServer}
            serverName={this.state.activeServerName}
          />
        );
      case "server_admin":
        return <ServerList />;
      case "user_admin":
        return <UserList />;
      case "group_admin":
        return <GroupList />;
      case "user_details":
        return <UserDetails />;
      case "public_key_admin":
        return <PublicKey />;
      default:
        return <div />;
    }
  };

  render() {
    const { classes, is_admin } = this.props;
    const { activePage } = this.state;

    return (
      <React.Fragment>
        <CssBaseline />
        <div className={classes.root}>
          <AppBar
            position="absolute"
            className={classNames(
              classes.appBar,
              this.state.open && classes.appBarShift
            )}
            color={is_admin ? "secondary" : "primary"}
          >
            <Toolbar
              disableGutters={!this.state.open}
              className={classes.toolbar}
            >
              <IconButton
                color="inherit"
                aria-label="Open drawer"
                onClick={this.handleDrawerOpen}
                className={classNames(
                  classes.menuButton,
                  this.state.open && classes.menuButtonHidden
                )}
              >
                <MenuIcon />
              </IconButton>
              <Typography
                variant="h6"
                color="inherit"
                noWrap
                className={classes.title}
              >
                FlowAuth
                {is_admin ? ": Admin Mode" : ""}
              </Typography>
              <IconButton
                id="user_details"
                color="inherit"
                onClick={() => this.setActivePage("user_details")}
              >
                <AccountCircleIcon />
              </IconButton>
              <IconButton
                id="logout"
                color="inherit"
                onClick={this.handleLogout}
              >
                <ExitToAppIcon />
              </IconButton>
            </Toolbar>
          </AppBar>
          <Drawer
            variant="permanent"
            classes={{
              paper: classNames(
                classes.drawerPaper,
                !this.state.open && classes.drawerPaperClose
              ),
            }}
            open={this.state.open}
          >
            <Divider />
            {is_admin && (
              <React.Fragment>
                <AdminMenu onClick={this.setActivePage.bind(this)} />
                <Divider />
              </React.Fragment>
            )}
            <UserServerList onClick={this.setServer.bind(this)} />
          </Drawer>
          <main className={classes.content}>
            <div className={classes.appBarSpacer} />
            <Typography variant="h4" gutterBottom>
              {this.page(activePage)}
            </Typography>
          </main>
        </div>
      </React.Fragment>
    );
  }
}

Dashboard.propTypes = {
  classes: PropTypes.object.isRequired,
};

export default withStyles(styles)(Dashboard);
